#requires -version 2.0
<#
LCDDisplay

PowerShell module that implements Matrix Orbital emulation on a networked LCD display such as 
the #twatch or my SNTPDisplay

David C. Moisan

Version 1.0

May 2012

Legal Information

This file is part of the SNTPDisplay project by David Moisan.
This file is based on the #twatch project (dangerousprototypes.com).
Code written by and for the SNTPDisplay project is released into the public domain, 

Originally written by Ian Lesnet <ian@dangerousprototypes.com>.
Modified by David Moisan <dmoisan@davidmoisan.org>.

To the extent possible under law, the author (David Moisan) has
waived all copyright and related or neighboring rights to this code. This
work is published from United States.

For details see: http://creativecommons.org/publicdomain/zero/1.0/.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

Installation

See the file about_lcddisplay_module for installation instructions and general usage
information.

Comment Based Help
Comment based help is implemented for all functions.

#>
function Find-MCHPIPDevice {
<#
.SYNOPSIS
Find-MCHPIPDevice

.DESCRIPTION
Sends a broadcast packet to UDP 30303 and waits for a response from
a Microchip device that uses the TCP/IP stack.

.OUTPUTS
Returns one object corresponding to the first device that respond to a discovery broadcast.
Fields are as follows.  All are strings.
IPAddr		IP Address of device responding
Hostname	Name of device responding
MACAddr 	MAC address of device responding, in string format.
ResetReason	Reason for last reset
#>

	param() # No parameters

	begin {
		Set-StrictMode -Version 2.0
	}

	process {
		$MICROCHIP_DISCOVERY_PORT = 30303
		$MICROCHIP_DISCOVERY_STRING = "Discovery: Who is out there?" 

		# ASCII Encoding to byte array object

		$encodebytes = New-Object System.Text.ASCIIEncoding

		# Device Discovery

		# Create socket and endpoints for broadcast receive and transmit

		$localmachineendpoint = new-object System.Net.IPEndPoint([System.Net.IPAddress]::Any,0)
		$mchpdiscoveryconnection = New-Object System.Net.Sockets.UDPClient($localmachineendpoint)
		$mchpdiscoveryremoteaddress = New-Object System.Net.IPEndpoint([System.Net.IPAddress]::Any,0)
		$mchpdiscoverybroadcastaddress = New-Object System.Net.IPEndPoint([System.Net.IPAddress]::Broadcast,$MICROCHIP_DISCOVERY_PORT)

		# Set receive and send timeout and enable broadcast

		$mchpdiscoveryconnection.Client.ReceiveTimeout = 30000 # 30 seconds
		$mchpdiscoveryconnection.Client.SendTimeout = 10000 # 10 seconds
		$mchpdiscoveryconnection.Client.EnableBroadcast = $true

		# Send discovery packet

		$discoverypacket = $encodebytes.GetBytes($MICROCHIP_DISCOVERY_STRING)
		$mchpsendstatus = $mchpdiscoveryconnection.Send($discoverypacket,$discoverypacket.Length,$mchpdiscoverybroadcastaddress)

		# Wait for response - 30 seconds ($receivetimeout/$receivedelay) 
		try {
			$receivetimeout = 30
			$receivedelay = 1
			$receivecount = [int]$receivetimeout / $receivedelay
			$receivedataflag = $false
			while (($receivecount -ge 0) -and (-not $receivedataflag)) {
				if ($mchpdiscoveryconnection.Available -gt 0) {
					$receivedataflag = $true
				}
				else {
					$receivecount--
					sleep $receivedelay
				}
			}
		}
		catch {
			Write-Error "Socket connection error: $_"
		}

		if ($receivedataflag) {
			$discoveryresultsreceivebuffer = $mchpdiscoveryconnection.Receive([ref]$mchpdiscoveryremoteaddress)

			$mchpaddress = $mchpdiscoveryremoteaddress.Address.IPAddressToString

			$discoverystring = $encodebytes.GetString($discoveryresultsreceivebuffer)
			$discoveryresults = $discoverystring -split "`n"
			
			if ($discoveryresults -ne $null) {
				New-Object PSObject -Property @{
					IPAddress = $mchpaddress 
					HostName = $discoveryresults[0] 
					MACaddr = $discoveryresults[1] 
					ResetReason = $discoveryresults[2] 
				} | write-output
			}
		} 
		$mchpdiscoveryconnection.Close() | Out-Null
		return
	}
}

function Connect-Lcd {
<#
.SYNOPSIS
connect-lcd

.DESCRIPTION
Connect to an IP LCD display implementing the Matrix Orbital protocol

.PARAMETER ipaddress
IP Address of LCD

.PARAMETER port
Port to connect to.  Default is 1337

.OUTPUTS
LCD Connection object [System.Net.Sockets.TcpClient] that represents the LCD device connected
#>

	param([Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="IP address of LCD")][string]$ipaddress,`
		[Parameter(Mandatory=$false,ValueFromPipeline=$false,HelpMessage="Port number")][int]$port = 1337)

	begin {
		Set-StrictMode -Version 2.0
	}

	process {
		$connection = New-Object System.Net.Sockets.TCPClient($ipaddress,$port)
		$connection | write-output
		return
	}
}

function Enter-LCDmode {
<#
.SYNOPSIS
enter-lcdmode

.DESCRIPTION
Send command to enter LCD mode

.PARAMETER connection
[System.Net.Sockets.TcpClient] Object representing the current connection

#>
	param ( `
		[Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="LCD Connection object")] `
		[System.Net.Sockets.Tcpclient]$connection `
	)

	begin {
		Set-StrictMode -Version 2.0
	}

	process { 
		$CMD_DISPLAY = 254
		$DISP_POS = 71
		$CMD_EnterLCDMode = $($CMD_DISPLAY, $DISP_POS, 1, 1)
		$socket = $connection.GetStream()
		$socket.Write($CMD_EnterLCDMode,0,$CMD_EnterLCDMode.Length) | out-null
		return
	}
}
function Clear-LCD {
<#
.SYNOPSIS
clear-lcd

.DESCRIPTION
Clear LCD screen

.PARAMETER connection
[System.Net.Sockets.TcpClient] Object representing the current connection

#>
	param(`
		[Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="LCD Connection object")] `
		[System.Net.Sockets.Tcpclient]$connection`
	)
	begin {
		Set-StrictMode -Version 2.0
	}

	process {
		$CMD_DISPLAY = 254
		$DISP_Clear = $CMD_DISPLAY, 88 # Clear Display
		$socket = $connection.GetStream()
		$socket.Write($DISP_Clear,0,$DISP_Clear.Length) | out-null
		return
	}
}

function Set-LCDbacklight {
<#
.SYNOPSIS
set-lcdbacklight

.DESCRIPTION
Turn LCD backlight on or off

.PARAMETER connection
[System.Net.Sockets.TcpClient] Object representing the current connection

.PARAMETER light
Backlight state:
$true  Backlight on
$false Backlight off
#>
	param(`
		[Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="LCD Connection object")] `
		[System.Net.Sockets.Tcpclient]$connection, `
		[Parameter(Mandatory=$false,ValueFromPipeline=$false,HelpMessage="Backlight state")] `
		[boolean]$light = $true `
	)

	process {
		Set-StrictMode -Version 2.0

		$CMD_DISPLAY = 254
		$DISP_LightON = $($CMD_DISPLAY,66,0) # Backlight on 
		$DISP_LightOFF = $($CMD_DISPLAY,70) # Backlight off

		$socket = $connection.GetStream()

		if ($light) {
			$socket.Write($DISP_LightON,0,$DISP_LightON.Length) | out-null
		}
		else {
			$socket.Write($DISP_LightOFF,0,$DISP_LightOFF.Length) | Out-Null
		}
		return
	}
}

function Show-NetworkonLCD {
<#
.SYNOPSIS
show-networkonlcd

.DESCRIPTION
Display network IP, gateway and DNS information on the LCD

.PARAMETER connection
[System.Net.Sockets.TcpClient] Object representing the current connection

#>
	param(`
		[Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="LCD Connection object")] `
		[System.Net.Sockets.Tcpclient]$connection`
	)
	begin {
		Set-StrictMode -Version 2.0
	}

	process {
		$CMD_DISPLAY = 254
		$DISP_Netcfg = $($CMD_DISPLAY, 253) # Display Network Configuration
		$socket = $connection.GetStream()
		$socket.Write($DISP_Netcfg,0,$DISP_Netcfg.Length) | out-null
		return
	}
}

function Set-LCDbrightness {
<#
.SYNOPSIS
set-lcdbrightness

.DESCRIPTION
Set LCD Brightness

.PARAMETER connection
[System.Net.Sockets.TcpClient] Object representing the current connection

.PARAMETER brightness
Desired brightness 
(default 255)
.NOTES
Not implemented in #Twatch or SNTPDisplay by default

#>
	param(`
		[Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="LCD Connection object")] `
		[System.Net.Sockets.Tcpclient]$connection,`
		[Parameter(Mandatory=$false,ValueFromPipeline=$false,HelpMessage="LCD Brightness")] `
		[byte]$brightness = 255 `
	)
	begin {
		Set-StrictMode -Version 2.0
	}
	process {
		$CMD_DISPLAY = 254
		$DISP_Bright = $($CMD_DISPLAY, 152, $brightness) # Display brightness, followed by desired level [1 byte].

		$socket = $connection.GetStream()
		$socket.Write($DISP_Bright,0,$DISP_Bright.Length) | out-null
		return
	}
}

function Set-LCDcustomchar {
<#
.SYNOPSIS
set-lcdcustomchar

.DESCRIPTION
Define LCD custom character by specifying an eight-byte bitmap and the character number to replace

.PARAMETER connection
[System.Net.Sockets.TcpClient] Object representing the current connection

.PARAMETER charnum
Character position to replace with custom bitmap.  Characters 0 through 7 can be 
replaced with a custom bitmap.

.PARAMETER bitmap
Eight-byte bitmap array that corresponds to a 5x8 character matrix.
See NOTES for full instructions.

.NOTES
The character bitmap is encoded as follows:

Byte      Bitmap
0         ***nnnnn
1         ***nnnnn
2         ***nnnnn
3         ***nnnnn
4         ***nnnnn
5         ***nnnnn
6         ***nnnnn
7         ***nnnnn

* = Ignored.
Only the five least significant bits are used to compose the character--the three higher bits are ignored.
See the HD44780 data sheet for more details.

#>
	param(`
		[Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="LCD Connection object")] `
		[System.Net.Sockets.Tcpclient]$connection,`
		[Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="Character number")] `
		[byte]$charnum, `
		[Parameter(Mandatory=$true,HelpMessage="Bitmap")][byte[]]$bitmap `
	)
	begin {
		Set-StrictMode -Version 2.0
	}

	process {
		$CMD_DISPLAY = 254
		$DISP_Custom = $($CMD_DISPLAY, 78, $charnum) + $bitmap # Custom character command:  followed by eight bytes defining the bitmap

		$socket = $connection.GetStream()
		$socket.Write($DISP_Custom,0,$DISP_Custom.Length) | out-null
		return
	}
}

function Set-LCDposition {
<#
.SYNOPSIS
set-lcdposition

.DESCRIPTION
Set LCD Cursor position

.PARAMETER connection
[System.Net.Sockets.TcpClient] Object representing the current connection

.PARAMETER column
LCD column

.PARAMETER row
LCD row

.NOTES
Because displays based on the HD44780 can be, and often are, of many different sizes, this function does not attempt to 
perform cursor position validation.  All cursor values are passed through to the firmware.
Valid positions for the #twatch are 1 through 20 columns by 1 through 4 rows.
#>
	param(`
		[Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="LCD Connection object")] `
		[System.Net.Sockets.Tcpclient]$connection, `
		[Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="LCD column position")] `
		[byte]$column, `
		[Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="LCD row position")] `
		[byte]$row `
	)
	begin {
		Set-StrictMode -Version 2.0
	}

	process {
		$CMD_DISPLAY = 254
		$DISP_POS = 71
		$DISP_Setpos = $($CMD_DISPLAY, $DISP_POS, $column, $row)
		$socket = $connection.GetStream()
		$socket.Write($DISP_Setpos,0,$DISP_Setpos.Length) | out-null
		return
	}
}

function set-lcdhome {
<#
.SYNOPSIS
set-lcdhome

.DESCRIPTION
Set the LCD cursor to home position

.PARAMETER connection
[System.Net.Sockets.TcpClient] Object representing the current connection

#>
	param(`
		[Parameter(Mandatory=$true,HelpMessage="LCD Connection object")] `
		[System.Net.Sockets.Tcpclient]$connection`
	)

	begin {
		Set-StrictMode -version 2.0
	}

	process {
		$CMD_DISPLAY = 254
		$DISP_Home = $($CMD_DISPLAY, 72) # Set cursor home

		$socket = $connection.GetStream()
		$socket.Write($DISP_Home,0,$DISP_Home.Length) | out-null
		return
	}
}

function set-lcdcursortype {
<#
.SYNOPSIS

set-lcdcursortype
.DESCRIPTION
Enable or disable the cursor and set it to block or underline

.PARAMETER connection
[System.Net.Sockets.TcpClient] Object representing the current connection

.PARAMETER underline
Specify underline cursor (default)

.PARAMETER block
Specify block cursor

.PARAMETER enable
Boolean value to enable ($true) or disable ($false) specified cursor
#>
	param(`
		[Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="LCD Connection object")] `
		[System.Net.Sockets.Tcpclient]$connection,`
		[Parameter(Mandatory=$false,ValueFromPipeline=$false,HelpMessage="Specify Underline cursor")] `
		[switch]$underline,`
		[Parameter(Mandatory=$false,ValueFromPipeline=$false,HelpMessage="Specify Block cursor")] `
		[switch]$block,`
		[Parameter(Mandatory=$false,ValueFromPipeline=$false,HelpMessage="Cursor enabled or disabled")] `
		[boolean]$enable = $false`
	)

	begin {
		Set-StrictMode -Version 2.0
	}

	process {
		$CMD_DISPLAY = 254
		$DISP_ULON = 74 # Underline cursor on
		$DISP_ULOFF = 75 # Underline cursor off
		$DISP_BlockON = 83 # Block cursor on
		$DISP_BlockOFF = 84 # Block cursor off

		# Set cursor based on parameters.  

		if ($block.isPresent) {
			if ($enable) {
				$cursordisplaystring = $($CMD_DISPLAY,$DISP_BlockON)
			}
			else { 
				$cursordisplaystring = $($CMD_DISPLAY,$DISP_BlockOFF)
			}
		}

		# If no underline cursor switch is present, assume it (it is the default).
		else {
			if ($enable) {
				$cursordisplaystring = $($CMD_DISPLAY,$DISP_ULON)
			}
			else {
				$cursordisplaystring = $($CMD_DISPLAY,$DISP_ULOFF)
			}
		}

		$socket = $connection.GetStream()
		$socket.Write($cursordisplaystring,0,$cursordisplaystring.Length) | out-null
		return
	}
}

function Write-LCD {
<#
.SYNOPSIS
write-lcd

.DESCRIPTION
Write a text string to LCD

.PARAMETER connection
[System.Net.Sockets.TcpClient] Object representing the current connection

#>
	param(`
		[Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="LCD Connection object")] `
		[System.Net.Sockets.Tcpclient]$connection,`
		[Parameter(Mandatory=$true,HelpMessage="Text to send to LCD")] `
		[string]$textstring`
	)
	begin {
		Set-StrictMode -version 2.0
	}

	process {
		$encodebytes = New-Object System.Text.ASCIIEncoding
		$socket = $connection.GetStream()
		$textstream = $encodebytes.GetBytes($textstring)
		$socket.Write($textstream,0,$textstream.Length) | out-null
		return
	}
}

function exit-lcdmode {
<#
.SYNOPSIS
exit-lcdmode

.DESCRIPTION
Send command to exit LCD emulation mode and return to SNTP Display (or #twatch).

.PARAMETER connection
[System.Net.Sockets.TcpClient] Object representing the current connection
#>
	param ([Parameter(Mandatory=$true,ValueFromPipeline=$false,HelpMessage="LCD Connection object")] `
	[System.Net.Sockets.Tcpclient]$connection `
	)

	begin {
		Set-StrictMode -Version 2.0
	}

	process {
		$CMD_DISPLAY = 254
		$CMD_ExitLCDMode = $($CMD_DISPLAY, 154) # Exit Matrix Orbital mode
		$socket = $connection.GetStream()
		$socket.Write($CMD_EXITLCDMode,0,$CMD_EXITLCDMode.Length) | out-null
		return
	}
}

function close-lcd {
<#
.SYNOPSIS
close-lcd

.DESCRIPTION
Close LCD connection

.PARAMETER connection
[System.Net.Sockets.TcpClient] Object representing the current connection

#>
	param([Parameter(Mandatory=$true,HelpMessage="LCD Connection Object")] `
	[System.Net.Sockets.Tcpclient]$connection)

	begin {
		Set-StrictMode -Version 2.0
	}

	process {
		$socket = $connection.GetStream()
		$socket.Close()
		$connection.Close()
		return
	}
}
## -------------------------------------------------------------------- ##

## Main Demo Code ##

function Start-LCDDemo {
<#
.SYNOPSIS
Start-LCDDemo

.DESCRIPTION
Demonstrates LCD functions.

.NOTES
This is a 2-minute demonstration that exercises all of the functions of the LCD display
#>
	begin {
		Set-StrictMode -version 2.0
	}

	process { # Get SNTP Display address
		Write-Host "Finding display..."

		$sntpdisplayaddress = Find-MCHPIPDevice

		if ($sntpdisplayaddress -eq $null) {
			Write-Error "No devices found!"
			return
		}
		else {
			Write-Host @"
SNTP Display found!

IPAddress:	$($sntpdisplayaddress.IPAddress)
MAC:		$($sntpdisplayaddress.MACAddr)
Reset:		$($sntpdisplayaddress.ResetReason)
"@
		}


		## Connect to the LCD

		$sntpdisplay = connect-lcd -ipaddress $sntpdisplayaddress.IPAddress

		# Enter LCD Mode. but not before displaying the clock for a few seconds

		sleep 7

		enter-lcdmode -connection $sntpdisplay
		clear-lcd -connection $sntpdisplay
		set-lcdhome -connection $sntpdisplay

		# Set brightness--does nothing (not supported on display)--but still need to test the command
		set-lcdbrightness -connection $sntpdisplay -brightness 255 

		set-lcdposition -connection $sntpdisplay -column 2 -row 2
		write-lcd -connection $sntpdisplay -textstring "SNTP Display Clock"
		sleep 3

		set-lcdposition -connection $sntpdisplay -column 3 -row 3
		write-lcd -connection $sntpdisplay -textstring "by David Moisan"
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "This is a demo of my"
		set-lcdposition -connection $sntpdisplay -row 2 -column 1
		write-lcd -connection $sntpdisplay "PowerShell modules" 
		set-lcdposition -connection $sntpdisplay -row 3 -column 1 
		write-lcd -connection $sntpdisplay -textstring "working with the"
		set-lcdposition -connection $sntpdisplay -row 4 -column 1
		write-lcd -connection $sntpdisplay -textstring "#twatch."
		sleep 4

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "I've converted mine"
		set-lcdposition -connection $sntpdisplay -column 1 -row 2
		write-lcd -connection $sntpdisplay -textstring "to an SNTP display."
		sleep 3

		clear-lcd -connection $sntpdisplay 
		write-lcd -connection $sntpdisplay -textstring "I liked the #twatch, but I've always wanted an SNTP display."
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "I also wanted the Matrix Orbital emulation of the #twatch firmware."
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "so that my servers"
		set-lcdposition -connection $sntpdisplay -row 2 -column 1 
		write-lcd -connection $sntpdisplay -textstring "can give me warning messages."
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "Because the #twatch doesn't have config memory,"
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "and because the flash on the Microchip  CPU is limited to 100 writes"
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "my SNTP display had to work with no configuration."
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "So it displays UTC,"
		sleep 3

		exit-lcdmode -connection $sntpdisplay
		sleep 5
		enter-lcdmode -connection $sntpdisplay

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "and uses the default NTP pool servers for time."
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "It's good enough for a display."
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "And UTC is fine for sysadmins, techs and hams."
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "The LCD emulation in the #twatch is fully supported."
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "Such as cursor"
		set-lcdposition -connection $sntpdisplay -row 2 -column 1
		write-lcd -connection $sntpdisplay -textstring "positioning:"
		sleep 2

		clear-lcd -connection $sntpdisplay
		set-lcdposition -connection $sntpdisplay -column 2 -row 1
		write-lcd -connection $sntpdisplay -textstring "Writing text..."
		sleep 1

		set-lcdposition -connection $sntpdisplay -column 8 -row 2
		write-lcd -connection $sntpdisplay -textstring "...here,"
		sleep 1

		set-lcdposition -connection $sntpdisplay -column 4 -row 3
		write-lcd -connection $sntpdisplay -textstring "here, "
		sleep 1

		set-lcdposition -connection $sntpdisplay -column 10 -row 4
		write-lcd -connection $sntpdisplay -textstring "and here!"
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "Block cursor:"
		set-lcdcursortype -connection $sntpdisplay -block -enable:$true
		sleep 3

		clear-lcd -connection $sntpdisplay
		set-lcdcursortype -connection $sntpdisplay -block -enable:$false
		write-lcd -connection $sntpdisplay -textstring "Underline cursor:"
		set-lcdcursortype -connection $sntpdisplay -underline -enable:$true
		sleep 3

		set-lcdcursortype -connection $sntpdisplay -underline -enable:$false

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "Turning the"
		set-lcdposition -connection $sntpdisplay -row 2 -column 1 
		write-lcd -connection $sntpdisplay -textstring "backlight off..."
		sleep 2

		set-lcdbacklight -connection $sntpdisplay -light:$false
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "and back on..."
		set-lcdbacklight -connection $sntpdisplay -light:$true
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "We can get the LCD's IP address."
		sleep 3

		show-networkonlcd -connection $sntpdisplay
		sleep 5

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "MCHP discovery is fully supported"
		sleep 3

		# Load custom characters

		# Left triangular arrow

		$leftarrow = [byte[]]$(0,1,3,7,3,1,0,0)

		# Right triangular arrow

		$rightarrow = [byte[]]$(0,4,6,7,6,4,0,0)

		# Solid square

		$square = [byte[]]$(0,0,15,15,15,15,0,0)

		# Solid circle

		$circle = [byte[]]$(0,0,6,15,15,6,0,0)

		# Up triangular arrow

		$uparrow = [byte[]]$(0,0,4,4,6,15,0,0)

		# Down triangular arrow

		$downarrow = [byte[]]$(0,0,15,6,4,4,0,0)

		# Hollow circle

		$hollowcircle = [byte[]]$(0,0,6,9,9,6,0,0)

		# Hollow square

		$hollowsquare = [byte[]]$(0,0,15,9,9,15,0,0)

		set-lcdcustomchar -connection $sntpdisplay -charnum 0 -bitmap $square
		set-lcdcustomchar -connection $sntpdisplay -charnum 1 -bitmap $leftarrow
		set-lcdcustomchar -connection $sntpdisplay -charnum 2 -bitmap $rightarrow
		set-lcdcustomchar -connection $sntpdisplay -charnum 3 -bitmap $uparrow
		set-lcdcustomchar -connection $sntpdisplay -charnum 4 -bitmap $downarrow
		set-lcdcustomchar -connection $sntpdisplay -charnum 5 -bitmap $circle
		set-lcdcustomchar -connection $sntpdisplay -charnum 6 -bitmap $hollowcircle
		set-lcdcustomchar -connection $sntpdisplay -charnum 7 -bitmap $hollowsquare

		clear-lcd -connection $sntpdisplay
		set-lcdposition -connection $sntpdisplay -column 3 -row 1
		write-lcd -connection $sntpdisplay -textstring "Custom characters"
		sleep 1
        
        set-lcdposition -connection $sntpdisplay -column 3 -row 3

		for ($i = 0;$i -lt 8;$i++) {
			write-lcd -connection $sntpdisplay -textstring "$([char][byte]$i) "
			sleep 0.75
		}

		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "This demo is written in PowerShell with functions   like this:"
		sleep 5

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "write-lcd"
		set-lcdposition -connection $sntpdisplay -row 2 -column 2
		write-lcd -connection $sntpdisplay -textstring "-connection `$lcd"
		set-lcdposition -connection $sntpdisplay -row 3 -column 2
		write-lcd -connection $sntpdisplay -textstring "-textstring"
		set-lcdposition -connection $sntpdisplay -row 4 -column 2
		write-lcd -connection $sntpdisplay -textstring "`"SERVER DOWN!`""
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "SERVER DOWN!"
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "That`'s all!"
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "I hope you find"
		set-lcdposition -connection $sntpdisplay -row 2 -column 1 
		write-lcd -connection $sntpdisplay -textstring "this useful!"
		sleep 3

		clear-lcd -connection $sntpdisplay
		write-lcd -connection $sntpdisplay -textstring "Thanks to Ian Lesnet for the original   #twatch!"
		sleep 5 

		clear-lcd -connection $sntpdisplay
		set-lcdposition -connection $sntpdisplay -column 3 -row 1
		write-lcd -connection $sntpdisplay -textstring "Creative Commons"
		set-lcdposition -connection $sntpdisplay -column 5 -row 2
		write-lcd -connection $sntpdisplay -textstring "(CC-BY) 2012"
		set-lcdposition -connection $sntpdisplay -column 1 -row 3
		write-lcd -connection $sntpdisplay -textstring "David Moisan, N1KGH"
		sleep 3
		
        exit-lcdmode -connection $sntpdisplay
		close-lcd -connection $sntpdisplay
	}
}