/*
 * This file is part of the SNTPDisplay project, written by David Moisan
 * Modified from #twatch project  (dangerousprototypes.com).
 *
 * NOTE: this code requires the Microchip TCP/IP stack, which is released under the Microchip license.
 *       Code written by the #twatch project is released into the public domain, 
 *		 The Microchip license still applies to the Microchip TCP/IP stack.
 *
 * Originally written by Ian Lesnet, Dangerous Prototypes
 *
 * The author (David Moisan, <dmoisan@davidmoisan.org> has waived all copyright 
 * and related or neighboring rights to this code. This
 * work is published from United States.
 *
 * For details see: http://creativecommons.org/publicdomain/zero/1.0/.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#include "TCPIPConfig.h"
#include "TCPIP Stack/TCPIP.h"
#include "HD44780.h" //include LCD display functions
#include "LCDTCPServer.h"

// #define DEBUG 1 // If defined, display debugging diagnostic characters on LCD 

// LCD TCP server return codes

#define SERVER_PORT	1337 			//set the server listening port
 
// Define the Matrix Orbital command set
// All of these command are first prefixed with MATRIX_ORBITAL_COMMAND
//
#define MATRIX_ORBITAL_COMMAND 254 	// 0xFE: Command prefix
#define BACKLIGHT_ON 66            	// 0x42: 1 parameter (minutes 00=forever)
#define BACKLIGHT_OFF 70			// 0x46: Backlight off
#define CLEAR 88					// 0x58: Display clear
#define HOME 72						// 0x48: Display home
#define POSITION 71					// 0x47: 2 parameters (col, row)
#define UNDERLINE_CURSOR_ON 74		// 0x4A: Underline cursor on
#define UNDERLINE_CURSOR_OFF 75		// 0x4B: Underline cursor off
#define BLOCK_CURSOR_ON 83			// 0x53: Block cursor on
#define BLOCK_CURSOR_OFF 84			// 0x54: Block cursor off
#define BACKLIGHT_BRIGHTNESS 152	// 0x98: 1 parameter (brightness)
#define CUSTOM_CHARACTER 78			// 0x4E: 9 parameters (character #, 8 byte bitmap)
#define NETWORK_PARAMETERS 253		// 0xFD: Announce IP and display on LCD
#define EXIT 154					// 0x9A: exit Matrix orbital mode and resume clock

unsigned char MatrixOrbitalTCPServer(void)
{
	BYTE i, workingCommand, ret=0;
	WORD w, w2;
	WORD msgcharindex=0;
	BYTE AppBuffer[64];
	WORD wMaxGet, wCurrentChunk;
	static TCP_SOCKET MySocket;
	
static enum _TCPServerState
	{
		SM_HOME = 0,
		SM_LISTENING,
	} TCPServerState = SM_HOME; //server status state machine

	static enum _LCDmode{
		TEXT=0,
		LCD_COMMAND,
		LCD_COMMAND_PARAMETERS,
	} LCDmode=TEXT; //matrix orbital command parser state machine
	
	static unsigned char parameters, parCnt, par[9]; //variables to store LCD incoming commands

	#if defined(DEBUG) // DEBUG:  Show LCD mode on display as 'M<mode>'
		LCD_WriteChar('M');
		LCD_WriteChar('0'+LCDmode);
	#endif

	//This state machine responds to incoming TCP packets and parses it for matrix orbital commands
	//and SNTP Display commands
	switch(TCPServerState)
	{
		case SM_HOME://start server listening
			MySocket = TCPOpen(0, TCP_OPEN_SERVER, SERVER_PORT, TCP_PURPOSE_GENERIC_TCP_SERVER); //setup server
			if(MySocket == INVALID_SOCKET) //if invalid, try again next time
				return RETNOINPUT;

			TCPServerState = SM_LISTENING; //server listening
			break; // End case SM_HOME

		case SM_LISTENING://service the TCP server
			if(!TCPIsConnected(MySocket)) //if no connections, break
				return RETNOINPUT;

			wMaxGet = TCPIsGetReady(MySocket);	//get TCP RX byte count

			// Process all bytes that we can
			// This is implemented as a loop, processing up to sizeof(AppBuffer) bytes at a time.  
			// This limits memory usage while maximizing performance.  Single byte Gets and Puts are a lot slower than multibyte GetArrays and PutArrays.
			wCurrentChunk = sizeof(AppBuffer);
			for(w = 0; w < wMaxGet; w += sizeof(AppBuffer))
			{
				if(w + sizeof(AppBuffer) > wMaxGet) //adjust for odd size arrays
					wCurrentChunk = wMaxGet - w;

				TCPGetArray(MySocket, AppBuffer, wCurrentChunk); //get data
				
				for(w2 = 0; w2 < wCurrentChunk; w2++)
				{
					i = AppBuffer[w2];

					//-------------------------------//
					//parser
					switch(LCDmode){ //this is a state machine that grabs the incoming commands one byte at a time
						case TEXT: 	// Default starting state is TEXT.
							if(i==MATRIX_ORBITAL_COMMAND){
								#if defined(DEBUG) // DEBUG:  Show 'C' on LCD for entry to command mode
									LCD_WriteChar('C');
								#endif				
								LCDmode=LCD_COMMAND; //next byte is actual command
								} 
							else { // Write char to the LCD.  No mode change.	
								LCD_WriteChar(i); 
							}							
							break; // End case TEXT
						
						case LCD_COMMAND:	//previous byte was 254, now get actual command
							parameters=0;  // Reset parameters count
							switch(i){	//switch on the command
								case BACKLIGHT_ON: 		//1 parameter (minutes 00=forever)
									LCD_Backlight(1);	//turn it on, we ignore the parameter
									LCDmode=LCD_COMMAND_PARAMETERS;//get one more byte before processing
									parameters=1;
									break; // End case BACKLIGHT_ON

								case BACKLIGHT_OFF:
									LCD_Backlight(0);	// Backlight off
									break; // End case BACKLIGHT_OFf
								
								case CLEAR:
									LCD_Clear();		// Clear LCD
									break; // End case CLEAR
								
								case HOME:
									LCD_Home();			// Home LCD
									break; // End case HOME
							
								case POSITION: //2 parameters (col, row)
									LCDmode=LCD_COMMAND_PARAMETERS; //Change mode:  
									parameters=2; // get 2 more bytes before processing
									break; // End case POSITION
						
								case UNDERLINE_CURSOR_ON:
									LCD_UnderlineCursor(1);
									break; // End case UNDERLINE_CURSOR_ON
							
								case UNDERLINE_CURSOR_OFF:
									LCD_UnderlineCursor(0);
									break; // End case UNDERLINE_CURSOR_OFF
								
								case BLOCK_CURSOR_ON:
									LCD_BlinkCursor(1);
									break; // End case BLOCK_CURSOR_ON
							
								case BLOCK_CURSOR_OFF:
									LCD_BlinkCursor(0);
									break; // End case BLOCK_CURSOR_OFF
							
								case BACKLIGHT_BRIGHTNESS://1 parameter (brightness)
									LCDmode=LCD_COMMAND_PARAMETERS; //Change mode , and get one more byte before processing
									parameters=1;
									break; // End case BACKLIGHT_BRIGHTNESS
							
								case CUSTOM_CHARACTER: //9 parameters (character #, 8 byte bitmap)
									LCDmode=LCD_COMMAND_PARAMETERS; //get 9 more bytes before processing
									parameters=9;
									break; // End case CUSTOM_CHARACTER

								case NETWORK_PARAMETERS:
									ret=RETIPANN;	// Return from TCP server and notify main() to announce IP address
									#if defined(DEBUG) // DEBUG:  Show 'D' to display IP address/announce request
										LCD_WriteChar('D');
									#endif
									break;	// No other mode change  End case NETWORK_PARAMETERS
															
								case EXIT: //restore time display mode  (custom command)
 									ret = RETEXITLCD;
									#if defined(DEBUG) // DEBUG:  Show 'X' to indicate exit LCD Display mode requested
										LCD_WriteChar('X');
										LCD_WriteChar('0'+ret);
									#endif
									
									break; // End case EXIT

								default: //error
									#if defined(DEBUG) // DEBUG: Show 'U' <command byte> to indicate unknown command
										LCD_WriteChar('U');
										LCD_WriteChar(i);
									#endif
									break; // End default
							}
						// Switch into command parameters state falls through here if no error

						if (LCDmode==LCD_COMMAND_PARAMETERS) { //if we have more bytes to get, store the command so we know what to do with them
							#if defined(DEBUG) // DEBUG:  Show 'P' to indicate LCD command parameter mode
								LCD_WriteChar('P');
							#endif
							workingCommand = i; //store command so we know how to deal with the parameters
							parCnt = 0; //reset the parameter counter
						}
						else { //all done with command
							LCDmode=TEXT; //next byte is text or begins new command				
							#if defined(DEBUG) // DEBUG:  Show 'T' to indicate entering text mode
								LCD_WriteChar('T');
							#endif
							}
						break;  // End case LCD_COMMAND

					case LCD_COMMAND_PARAMETERS:
						
						par[parCnt]=i;//store each parameter in a buffer
						parCnt++;
						if(parCnt<parameters) break; //if not all parameters yet, quit

						#if defined(DEBUG) // DEBUG:  Show 'p<parameters>' to LCD
							LCD_WriteChar('p');
							LCD_WriteChar('0'+parameters);
						#endif

						switch(workingCommand){ //got all the parameters, now process the command
							case BACKLIGHT_ON: //1 parameter (minutes 00=forever)
								 //ignore value
								break; // End case BACKLIGHT_ON

							case POSITION: //2 parameters (col, row)
								i=( ((par[1]-1)*20) + par[0] ); //convert to 20x4 layout (used defined lines, add rows...)
								LCD_CursorPosition(i);
								if(i==1) ret=RETSTARTLCD; //return 1 to stop twatch...
								break; // End case POSITION

							case BACKLIGHT_BRIGHTNESS://1 parameter (brightness)
								break; //do nothing, no brightness control (yet), use PWM on pin...

							case CUSTOM_CHARACTER: 			//9 parameters (character #, 8 byte bitmap)
								LCD_WriteCGRAM(par[0]);		//write character address
								for(i=1; i<9; i++) LCD_WriteRAM(par[i]); //send 8 bitmap bytes
								LCD_ExitCGRAM();
								#if defined(DEBUG) // Show '@' on LCD to indicate custom character loaded
									LCD_WriteChar('@');
								#endif
								break; // End case CUSTOM_CHARACTER

							default: //error
								#if defined(DEBUG) // DEBUG:  Display 'u' <parcnt> to indicate unknown command with parameters
									LCD_WriteChar('u');
									LCD_WriteChar('0'+parCnt);
								#endif
									
								break; // End default
						
						}
						LCDmode=TEXT; //end of command with parameters, next byte is text or new command...
						#if defined(DEBUG) // DEBUG:  Display 't' to indicate enter text mode from command with parameters
							LCD_WriteChar('t');
						#endif
						break;  // End case	LCD_COMMAND_PARAMETERS					
					}
				}
				//------------------------------//
			}
			break;  // End case SM_LISTENING_STATE
			}
	return ret; //return status info 
}

