/*
 * This file is part of the sntpdisplay project, written by David Moisan
 * Modified from #twatch project  (dangerousprototypes.com).
 *
 * NOTE: this code requires the Microchip TCP/IP stack, which is released under the Microchip license.
 *       Code written by the #twatch project is released into the public domain, 
 *		 The Microchip license still applies to the Microchip TCP/IP stack.
 *
 * The hardware was conceived and software originally written by Ian Lesnet, Dangerous Prototypes
 *
 * The author (David Moisan, <dmoisan@davidmoisan.org> has waived all copyright 
 * and related or neighboring rights to this code. This
 * work is published from United States.
 *
 * For details see: http://creativecommons.org/publicdomain/zero/1.0/.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
#include "TCPIPConfig.h"
#include "TCPIP Stack/TCPIP.h"
#include "HD44780.h" //include LCD display functions

#define TIMEDISPLAYLINE1 21  // Cursor position for first line of display
#define TIMEDISPLAYLINE2 41  // Cursor position for second line of display



void SendSNTPSecondsToClockDisplay(unsigned long unixepochtime) {

	// Converts SNTP Unix epoch seconds to displayable time

	// The following code to convert a Unix time epoch to time and date elements
	// is taken from Maxim AppNote #517:
	// http://www.maxim-ic.com/app-notes/index.mvp/id/517 
	

	// this array represents the number of days in one non-leap year at 
	//    the beginning of each month
 

	static unsigned int rom DaysToMonth[13] = {
   		0,31,59,90,120,151,181,212,243,273,304,334,365
		};

	unsigned int hour;
   	unsigned int day;
   	unsigned int minute;
   	unsigned int second;
   	unsigned int month;
   	unsigned int year;
     
   	unsigned long whole_minutes;
   	unsigned long whole_hours;
   	unsigned long whole_days;
   	unsigned long whole_days_since_1968;
   	unsigned long leap_year_periods;
   	unsigned long days_since_current_lyear;
   	unsigned long whole_years;
   	unsigned long days_since_first_of_year;
   	unsigned long days_to_month;
   	unsigned long day_of_week;

	// Given the unixepochtime, convert it to its components
       
   	whole_minutes = unixepochtime / 60;
   	second = unixepochtime - (60 * whole_minutes);		    // leftover seconds

   	whole_hours  = whole_minutes / 60;
   	minute = whole_minutes - (60 * whole_hours);            // leftover minutes

   	whole_days   = whole_hours / 24;
   	hour         = whole_hours - (24 * whole_days);         // leftover hours
        
   	whole_days_since_1968 = whole_days + 365 + 366;
   	leap_year_periods = whole_days_since_1968 / ((4 * 365) + 1);

  	days_since_current_lyear = whole_days_since_1968 % ((4 * 365) + 1);
        
   	// if days are after a current leap year then add a leap year period
   
   	if ((days_since_current_lyear >= (31 + 29))) {
      	leap_year_periods++;
   		}
   	whole_years = (whole_days_since_1968 - leap_year_periods) / 365;
	// Fixed to display day of year properly so January 1st displays as "Day 1"
   	days_since_first_of_year = whole_days_since_1968 - (whole_years * 365) - leap_year_periods + 1;

   	if ((days_since_current_lyear <= 365) && (days_since_current_lyear >= 60)) {
    	days_since_first_of_year++;
   		}
   	year = whole_years + 68;        

   	// 	setup for a search for what month it is based on how many days have passed
   	//  within the current year

   	month = 13;
   	days_to_month = 366;
//		Fixed to eliminate "Feb 0" problem where the last day of a month was displayed as the "0th" day of the next month
//		Changed comparison to be less than or equal
   	while (days_since_first_of_year <= days_to_month) {
    	month--;
       	days_to_month = DaysToMonth[month-1];
		if ((month > 2) && ((year % 4) == 0)) {
           days_to_month++;
			}
        }

	day = days_since_first_of_year - days_to_month;


 	day_of_week = (whole_days  + 4) % 7;

	// Display code
	//
	// Displays clock on LCD
	// Clock time comes from the tm struct we've created
	//
	// Display format:
	// Starting on character 21 (Line 2)
	//
	// Line 2: Day dyr hh:mm:ss UTC
	// Line 3: DAY MON dd, yyyy
	//
	// dyr       Day of year
	// hh:mm:ss  Hours, min, sec
	// UTC       Time zone
	//
	// DAY       Day abbreviation
	// MON       Month abbreviation
	// dd        Day of month
	// yyyy      Year
	//


	// First, set the cursor to home
	
	LCD_Home();
		
	// Line 1 Day of year

	LCD_CursorPosition(TIMEDISPLAYLINE1);
	LCD_WriteString("Day");
	LCD_CursorPosition(TIMEDISPLAYLINE1+4);
	LCD_WriteLeadingSpaceNumber(days_since_first_of_year,3);

	// Display the time, starting with hours
	// Note that 24-hour format is used
	LCD_CursorPosition(TIMEDISPLAYLINE1+8);
	LCD_WriteLeadingZeroNumber(hour, 2);
	LCD_CursorPosition(TIMEDISPLAYLINE1+10);
	LCD_WriteChar(':');

	// Display the minutes
	
	LCD_CursorPosition(TIMEDISPLAYLINE1+11);
	LCD_WriteLeadingZeroNumber(minute, 2);
	LCD_CursorPosition(TIMEDISPLAYLINE1+13);	
	LCD_WriteChar(':');

	// Display the seconds

	LCD_CursorPosition(TIMEDISPLAYLINE1+14);
	LCD_WriteLeadingZeroNumber(second, 2);
	
	// Display "UTC"

	LCD_CursorPosition(TIMEDISPLAYLINE1+17);
	LCD_WriteString("UTC");
	
	// Line 1 Display complete

	// Line 2
	
	// Display the day of week string
	LCD_CursorPosition(TIMEDISPLAYLINE2+2);
	switch (day_of_week) {
		case 0: 
			LCD_WriteString("Sun");
			break;
		case 1:
			LCD_WriteString("Mon");
			break;
		case 2:
			LCD_WriteString("Tue");				
			break;
		case 3:
			LCD_WriteString("Wed");
			break;
		case 4:
			LCD_WriteString("Thu");
			break;
		case 5:
			LCD_WriteString("Fri");
			break;
		case 6:
			LCD_WriteString("Sat");
			break;
		
		default:
			LCD_WriteString("Err");
			break;
	}
	
	// Display month string
	LCD_CursorPosition(TIMEDISPLAYLINE2+6);
	
	switch (month) {
		case 1:
			LCD_WriteString("Jan");
			break;
		case 2:
			LCD_WriteString("Feb");
			break;
		case 3:
			LCD_WriteString("Mar");
			break;
		case 4:
			LCD_WriteString("Apr");
			break;
		case 5:
			LCD_WriteString("May");
			break;
		case 6:
			LCD_WriteString("Jun");
			break;
		case 7:
			LCD_WriteString("Jul");
			break;
		case 8:
			LCD_WriteString("Aug");
			break;
		case 9:
			LCD_WriteString("Sep");
			break;
		case 10:
			LCD_WriteString("Oct");
			break;
		case 11:
			LCD_WriteString("Nov");
			break;
		case 12:
			LCD_WriteString("Dec");
			break;

		default:
			LCD_WriteString("Err");
			break;
	}	
	
	// Display day
	LCD_CursorPosition(TIMEDISPLAYLINE2+10);
	LCD_WriteLeadingSpaceNumber(day, 2);
	LCD_CursorPosition(TIMEDISPLAYLINE2+12);
	LCD_WriteString(", ");

	// Display year.  Four digits 
	LCD_CursorPosition(TIMEDISPLAYLINE2+14);
	LCD_WriteLeadingZeroNumber((year+1900), 4);
	
	// Line 2 complete

	// Display complete
	
	return;
	}
